#!/usr/bin/env python3
import os
import subprocess
import sys
from pathlib import Path
from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class StaticSiteBuildHook(BuildHookInterface):
    PLUGIN_NAME = "static_site"

    def initialize(self, version, build_data):
        """Run before the build process starts"""
        script_dir = Path(__file__).parent.absolute()

        os.chdir(script_dir)

        try:
            # Check if node is installed
            try:
                subprocess.run(
                    ["node", "--version"], check=True, capture_output=True, text=True
                )
            except (subprocess.CalledProcessError, FileNotFoundError):
                raise RuntimeError(
                    "Node.js is not installed or not in PATH. Please install Node.js to build the static site."
                )

            print("Running npm install...")
            result = subprocess.run(
                ["npm", "install", "--legacy-peer-deps"],
                check=True,
                capture_output=True,
                text=True,
            )
            print(result.stdout)

            print("Running npm build...")
            result = subprocess.run(
                ["npm", "run", "build"], check=True, capture_output=True, text=True
            )
            print(result.stdout)

            precompiled_dir = script_dir / "precompiled"
            if not precompiled_dir.exists():
                print("Error: precompiled directory not found after build")
                raise RuntimeError("precompiled directory not found after build")

            print("Static site build completed successfully")

        except subprocess.CalledProcessError as e:
            print(f"Error running npm command: {e.stderr}", file=sys.stderr)
            raise RuntimeError(f"npm command failed: {e.stderr}")
        except Exception as e:
            print(f"Unexpected error: {str(e)}", file=sys.stderr)
            raise



from typing import Any, Literal, Sequence, cast, NamedTuple

FilterOperator = Literal[
    "=",
    "!=",
    "<",
    "<=",
    ">",
    ">=",
    "in",
    "not in",
    "contains",
    "includes",
    "includes",
    "includes any",
    "includes all",
]


class Filter(NamedTuple):
    column: str
    operator: FilterOperator
    value: Any


InputFilters = Sequence[Filter] | Sequence[Sequence[Filter]]
NormalizedFilters = list[list[Filter]]


def normalize_filters(filters: InputFilters | None) -> NormalizedFilters:
    """
    Normalize filters to the standard format of a list of lists of filters, where each
    inner list represents a conjunction (AND) of filters and the outer list represents
    a disjunction (OR) of the inner conjunctions.

    We make a special case for the empty filter list, which is treated as no filters.
    Typically, a disjunction of no expressions should be false, but the user likely
    intends to apply no filters, and [] is easier to work with than [[]].
    """

    if filters is None or len(filters) == 0:
        return []
    elif all(isinstance(f, Filter) for f in filters):
        filters = cast(Sequence[Filter], filters)
        return [list(filters)]
    else:
        filters = cast(Sequence[Sequence[Filter]], filters)
        return [list(f) for f in filters]



from datarepo.core.catalog.catalog import Catalog, Database
from datarepo.core.tables.deltalake_table import DeltalakeTable
from datarepo.core.tables.metadata import TableProtocol
import json
import tempfile
import subprocess
from pathlib import Path
import os
import sys
import logging
import shutil
import importlib.resources

# Configure logging
logging.basicConfig(level=logging.INFO, format="%(message)s", stream=sys.stderr)
logger = logging.getLogger(__name__)


def export_table(name: str, table: TableProtocol):
    table_info = table.table_metadata

    schema = table.get_schema()
    partitions, columns = schema.partitions, schema.columns

    return {
        "name": name,
        "description": table_info.description,
        "partitions": partitions,
        "columns": columns,
        "selected_columns": table_info.docs_args.get("columns", None),
        "supports_sql_filter": isinstance(table, DeltalakeTable),
        "table_type": table_info.table_type,
        "latency_info": table_info.latency_info,
        "example_notebook": table_info.example_notebook,
        "data_input": table_info.data_input,
    }


def export_database(name: str, database: Database):
    tables = []

    for key, table in sorted(database.get_tables().items()):
        tables.append(export_table(key, table))

    return {
        "name": name,
        "tables": tables,
    }


def export_catalog(name: str, catalog: Catalog):
    return {
        "name": name,
        "databases": [export_database(key, catalog.db(key)) for key in catalog.dbs()],
    }


def export_datarepo(catalogs: list[tuple[str, Catalog]]) -> dict:
    return {
        "catalogs": [export_catalog(name, catalog) for name, catalog in catalogs],
    }


def export_and_generate_site(
    catalogs: list[tuple[str, Catalog]], output_dir: str = "./dist"
):
    """
    Export the catalog to a JSON file and copy the precompiled static site files.

    Args:
        catalogs: List of (name, catalog) tuples to export
        output_dir: Directory where the static site will be generated (default: "./dist")
    """
    # Export catalog to a JSON file
    catalog_data = export_datarepo(catalogs)

    output_path = Path(output_dir)
    # Remove output directory if it exists to ensure idempotency
    if output_path.exists():
        shutil.rmtree(output_path)
    output_path.mkdir(parents=True, exist_ok=True)

    with open(output_path / "data.json", "w") as f:
        json.dump(catalog_data, f)

    project_root = Path(__file__).parent
    precompiled_dir = project_root / "static_site" / "precompiled"

    logger.info(f"Copying precompiled directory {precompiled_dir} to {output_path}")
    if not precompiled_dir.exists():
        raise FileNotFoundError(
            f"Could not find precompiled directory. Make sure you're running from the project root or the package is properly installed."
        )

    shutil.copytree(precompiled_dir, output_path, dirs_exist_ok=True)



from datarepo.core import Catalog, NlkDataFrame

__all__ = ["NlkDataFrame", "Catalog"]


def repl():
    """
        Starts an interactive python session with some of the datarepo imports.
    Allows for quick testing and inspection of data accessible via the datarepo
        client.
    """

    import IPython
    import polars as pl

    from datarepo.core import Catalog, Filter, NlkDataFrame

    print(
        r"""
------------------------------------------------

Welcome to
     __                     _       _
  /\ \ \___ _   _ _ __ __ _| | __ _| | _____
 /  \/ / _ \ | | | '__/ _` | |/ _` | |/ / _ \
/ /\  /  __/ |_| | | | (_| | | (_| |   <  __/
\_\ \/ \___|\__,_|_|  \__,_|_|\__,_|_|\_\___|
------------------------------------------------

"""
    )

    IPython.start_ipython(
        colors="neutral",
        display_banner=False,
        user_ns={
            "Catalog": Catalog,
            "NlkDataFrame": NlkDataFrame,
            "Filter": Filter,
            "pl": pl,
        },
        argv=[],
    )



from datarepo.core.tables.decorator import table
from datarepo.core.tables.deltalake_table import (
    DeltaCacheOptions,
    DeltalakeTable,
)
from datarepo.core.tables.metadata import TableMetadata, TableProtocol, TableSchema
from datarepo.core.tables.parquet_table import ParquetTable
from datarepo.core.tables.util import (
    DeltaRoapiOptions,
    Filter,
    Partition,
    PartitioningScheme,
    RoapiOptions,
)

__all__ = [
    "DeltalakeTable",
    "DeltaCacheOptions",
    "ParquetTable",
    "PartitioningScheme",
    "Filter",
    "Partition",
    "table",
    "TableMetadata",
    "TableProtocol",
    "TableSchema",
    "DeltaRoapiOptions",
    "RoapiOptions",
]



import sys
from typing import Any

from datarepo.core import DeltalakeTable, ParquetTable
from datarepo.core.catalog.catalog import Catalog
from datarepo.core.tables.metadata import RoapiOptions, TableProtocol
from datarepo.core.tables.util import (
    DeltaRoapiOptions,
    Filter,
    PartitioningScheme,
)


def export_to_roapi_table(name: str, table: TableProtocol) -> dict[str, Any] | None:
    if isinstance(table, ParquetTable):
        return _export_parquet_table(name, table)
    elif isinstance(table, DeltalakeTable):
        return _export_deltalake_table(name, table)
    else:
        print(
            f"{name}: Only parquet and deltalake tables are supported for roapi export.",
            file=sys.stderr,
        )
        return None


def export_to_roapi_tables(catalog: Catalog) -> list[dict[str, Any]]:
    """
    Exports a list of roapi tables that can be injected into a roapi
    configuration yaml under the `tables` key.
    """
    roapi_tables = []
    for db_name in catalog.dbs():
        db = catalog.db(db_name)
        tables = db.get_tables(show_deprecated=True)

        for tbl_name, table in tables.items():
            roapi_table_name = f"{db_name}_{getattr(table, 'name', tbl_name)}"
            if (
                roapi_table := export_to_roapi_table(roapi_table_name, table)
            ) is not None:
                roapi_tables.append(roapi_table)

    return roapi_tables


def _export_parquet_table(name: str, table: ParquetTable) -> dict[str, Any] | None:
    roapi_opts = table.table_metadata.roapi_opts or RoapiOptions()

    if roapi_opts.disable:
        return None

    if table.partitioning_scheme != PartitioningScheme.HIVE:
        print(
            f"{table.name}: Only hive-partitioned parquet tables are supported by roapi.",
            file=sys.stderr,
        )
        return None

    table_info = table.table_metadata

    filters = table_info.docs_args.get("filters", [])
    if not filters:
        return {
            "name": roapi_opts.override_name or name,
            "uri": table.uri,
            "option": {
                "format": "parquet",
                "use_memory_table": roapi_opts.use_memory_table,
            },
        }

    elif not (filters and all(isinstance(x, Filter) for x in filters)):
        print(
            f"{table.name}: Must have docs filters defined to infer partition columns.",
            file=sys.stderr,
        )
        return None

    partition_cols = []
    for filter in filters:
        filter_name = filter.column
        data_type = py_type_to_roapi(type(filter.value))

        # Roapi can treat YYYY-MM-DD partition columns as "Date32"
        data_type = "Date32" if filter_name == "date" else data_type

        partition_cols.append(
            {
                "name": filter_name,
                "data_type": data_type,
            }
        )

    schema_from_file = table.build_file_fragment(filters)

    table_config = {
        "name": roapi_opts.override_name or name,
        "uri": table.uri,
        "option": {
            "format": "parquet",
            "use_memory_table": roapi_opts.use_memory_table,
        },
        "partition_columns": partition_cols,
        "schema_from_files": [schema_from_file],
    }

    return _with_reload_interval(table_config, roapi_opts)


def _export_deltalake_table(name: str, table: DeltalakeTable) -> dict[str, Any] | None:
    roapi_opts = table.table_metadata.roapi_opts or DeltaRoapiOptions()

    if roapi_opts.disable:
        return None

    table_config = {
        "name": roapi_opts.override_name or name,
        "uri": table.uri,
        "option": {
            "format": "delta",
            "use_memory_table": roapi_opts.use_memory_table,
        },
    }

    return _with_reload_interval(table_config, roapi_opts)


def _with_reload_interval(
    table_config: dict[str, Any], roapi_opts: RoapiOptions
) -> dict[str, Any]:
    if roapi_opts.reload_interval_seconds is not None:
        table_config["reload_interval"] = {
            "secs": roapi_opts.reload_interval_seconds,
            "nanos": 0,
        }
    return table_config


def py_type_to_roapi(py_type: type) -> str:
    return {
        int: "Int64",
        str: "Utf8",
        bool: "Boolean",
        float: "Float64",
    }[py_type]



from dataclasses import dataclass
from typing import Any, Dict, Protocol, TypedDict

from datarepo.core.dataframe.frame import NlkDataFrame
from datarepo.core.tables.util import RoapiOptions


@dataclass
class TableMetadata:
    """
    Information about a table used for documentation generation.
    """

    table_type: str
    description: str
    docs_args: Dict[str, Any]
    latency_info: str | None = None
    example_notebook: str | None = None
    data_input: str | None = None
    is_deprecated: bool = False
    roapi_opts: RoapiOptions | None = None


class TablePartition(TypedDict):
    column_name: str
    type_annotation: str
    value: Any


class TableColumn(TypedDict):
    column: str
    type: str
    readonly: bool
    filter_only: bool
    has_stats: bool


@dataclass
class TableSchema:
    partitions: list[TablePartition]
    columns: list[TableColumn]


class TableProtocol(Protocol):
    # Properties used to generate the web catalog & roapi config
    table_metadata: TableMetadata

    def __call__(self, **kwargs: Dict[str, Any]) -> NlkDataFrame: ...

    def get_schema(self) -> TableSchema:
        """
        Returns the schema of the table, used to generate the web catalog.
        """
        ...



# mypy: disable-error-code=override

import concurrent.futures
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
import os
from typing import Any, TypeAlias
import warnings

import boto3
from deltalake import DeltaTable, QueryBuilder
from deltalake.warnings import ExperimentalWarning
import polars as pl
import pyarrow as pa

from datarepo.core.dataframe import NlkDataFrame
from datarepo.core.tables.filters import InputFilters, normalize_filters
from datarepo.core.tables.metadata import (
    TableMetadata,
    TableProtocol,
    TableSchema,
)
from datarepo.core.tables.util import (
    DeltaRoapiOptions,
    Filter,
    RoapiOptions,
    filters_to_sql_predicate,
    get_storage_options,
)

READ_PARQUET_RETRY_COUNT = 10
DEFAULT_TIMEOUT = "150s"

DeltaInputFilters: TypeAlias = InputFilters | str


@dataclass
class DeltaCacheOptions:
    # Path to the directory where files are cached. This can be the same for all tables
    # since the S3 table prefix is included in the cached paths.
    file_cache_path: str
    # Duration for which the _last_checkpoint file is cached. A longer duration means
    # we can reuse cached checkpoints for longer, which could improve loading performance
    # from not downloading new checkpoint parquets. However, we also need to load more
    # individual transaction jsons, so this should not be set too high.
    # A reasonable value for a table with frequent updates (e.g. binned spikes) is 30m.
    file_cache_last_checkpoint_valid_duration: str | None = None

    def to_storage_options(self) -> dict[str, Any]:
        opts = {
            "file_cache_path": os.path.expanduser(self.file_cache_path),
        }
        if self.file_cache_last_checkpoint_valid_duration is not None:
            opts["file_cache_last_checkpoint_valid_duration"] = (
                self.file_cache_last_checkpoint_valid_duration
            )
        return opts


class DeltalakeTable(TableProtocol):
    def __init__(
        self,
        name: str,
        uri: str,
        schema: pa.Schema,
        description: str = "",
        docs_filters: list[Filter] = [],
        docs_columns: list[str] | None = None,
        roapi_opts: RoapiOptions | None = None,
        unique_columns: list[str] | None = None,
        table_metadata_args: dict[str, Any] | None = None,
        stats_cols: list[str] | None = None,
        extra_cols: list[tuple[pl.Expr, str]] | None = None,
    ):
        self.name = name
        self.uri = uri
        self.schema = schema
        self.unique_columns = unique_columns
        self.stats_cols = stats_cols or []
        self.extra_cols = extra_cols or []

        self.table_metadata = TableMetadata(
            table_type="DELTA_LAKE",
            description=description,
            docs_args={"filters": docs_filters, "columns": docs_columns},
            roapi_opts=roapi_opts or DeltaRoapiOptions(),
            **(table_metadata_args or {}),
        )

    def get_schema(self) -> TableSchema:
        dt = self.delta_table()
        schema = self.schema
        partition_cols = dt.metadata().partition_columns
        filters = {
            f.column: f.value
            for f in self.table_metadata.docs_args.get("filters", [])
            if isinstance(f, Filter)
        }
        partitions = [
            {
                "column_name": col,
                "type_annotation": str(schema.field(col).type),
                "value": filters.get(col),
            }
            for col in partition_cols
        ]
        columns = [
            {
                "name": name,
                "type": str(schema.field(name).type),
                "has_stats": name in partition_cols or name in self.stats_cols,
            }
            for name in schema.names
        ]
        columns += [
            {
                "name": expr.meta.output_name(),
                "type": expr_type,
                "readonly": True,
            }
            for expr, expr_type in self.extra_cols
        ]
        return TableSchema(
            partitions=partitions,
            columns=columns,
        )

    def __call__(
        self,
        filters: DeltaInputFilters | None = None,
        columns: list[str] | None = None,
        boto3_session: boto3.Session | None = None,
        endpoint_url: str | None = None,
        timeout: str | None = None,
        cache_options: DeltaCacheOptions | None = None,
        **kwargs: Any,
    ) -> NlkDataFrame:
        storage_options = {
            "timeout": timeout or DEFAULT_TIMEOUT,
            **get_storage_options(
                boto3_session=boto3_session, endpoint_url=endpoint_url
            ),
        }
        if cache_options is not None:
            storage_options = {
                **storage_options,
                **cache_options.to_storage_options(),
            }
        dt = self.delta_table(storage_options=storage_options)

        return self.construct_df(dt=dt, filters=filters, columns=columns)

    def construct_df(
        self,
        dt: DeltaTable,
        filters: DeltaInputFilters | None = None,
        columns: list[str] | None = None,
    ) -> NlkDataFrame:
        # Use schema defined on this table, the physical schema in deltalake metadata might be different
        schema = self.schema

        predicate_str = datafusion_predicate_from_filters(schema, filters)

        # These should not be read because they don't exist in the delta table
        extra_col_exprs = [expr for expr, _ in self.extra_cols]
        extra_column_names = set(expr.meta.output_name() for expr in extra_col_exprs)

        columns_to_read = None
        unique_column_names = set(self.unique_columns or [])
        if columns:
            columns_to_read = list(
                (set(columns) | unique_column_names) - extra_column_names
            )

        # TODO(peter): consider a sql builder for more complex queries?
        select_cols = (
            ", ".join([f'"{col}"' for col in columns_to_read])
            if columns_to_read
            else "*"
        )
        condition = f"WHERE {predicate_str}" if predicate_str else ""
        query_string = f"""
            SELECT {select_cols}
            FROM "{self.name}"
            {condition}
        """
        with warnings.catch_warnings():
            # Ignore ExperimentalWarning emitted from QueryBuilder
            warnings.filterwarnings("ignore", category=ExperimentalWarning)
            batches = (
                QueryBuilder().register(self.name, dt).execute(query_string).fetchall()
            )

        # Since we might cast unique string columns to categoricals, use a string cache to
        # improve performance when combining multiple dataframes
        with pl.StringCache():
            if batches:
                frame = pl.from_arrow(batches, rechunk=False)
                frame = _normalize_df(frame, self.schema, columns=columns_to_read)
            else:
                # If dataset is empty, the returned dataframe will have no columns
                frame = _empty_normalized_df(schema)

            if self.extra_cols:
                frame = frame.with_columns(extra_col_exprs)

            if self.unique_columns:
                # Cast unique string columns to categoricals first
                # In some cases, this reduces peak memory usage up to 50%
                curr_schema = frame.schema
                cat_schema = {
                    col: pl.Categorical
                    for col in curr_schema
                    if curr_schema[col] == pl.String
                }
                frame = (
                    frame.cast(cat_schema)
                    .unique(subset=self.unique_columns, maintain_order=True)
                    .cast(curr_schema)
                )

        if columns:
            frame = frame.select(columns)

        return frame.lazy()

    def delta_table(
        self, storage_options: dict[str, Any] | None = None, version: int | None = None
    ) -> DeltaTable:
        return DeltaTable(
            table_uri=self.uri, storage_options=storage_options, version=version
        )


def fetch_df_by_partition(
    dt: DeltaTable,
    partition: list[tuple[str, str, Any]],
    schema: pa.Schema,
    storage_options: dict[str, Any] | None = None,
) -> pl.DataFrame:
    """
    The native delta-rs read has slower performance. The difference comes from the dataset in
    dt.to_pyarrow_dataset() - this dataset is slower than the one manually constructed below.

    pyarrow.dataset.Dataset.to_table() reading was also tried, however this also resulted
    in slower performance for larger datasets. For example, reading P1 webgrid data on 2024-05-21
    (2.1M+ rows) takes ~20+ seconds.

    The hypothesized reason(s) for the slow read performance of pyarrow.dataset.Dataset.to_table()
    are
    1. Additional step of metadata loading
    2. Lazy loading which could lead to longer load times
    Source: https://github.com/apache/arrow/issues/35332

    When using pyarrow's parquet read_table(), which will eagerly read the entire dataset, the
    read for the P1 webgrid data on 2024-05-21 (2.1M+ rows) takes ~7-10 seconds.
    Thus, it makes sense to use read_table() since we have a predefined list of files to be loaded
    from dt.files

    """
    files = dt.files(partition_filters=partition)

    files = [os.path.join(dt.table_uri, file) for file in files]

    if not files:
        # NOTE: polars has a bug where with_columns(...) on an empty dataframe with no columns will add a row
        # for all new columns. Workaround by adding the columns before normalizing
        return _empty_normalized_df(schema)

    return fetch_dfs_by_paths(
        files=files, schema=schema, storage_options=storage_options
    )


def fetch_dfs_by_paths(
    files: list[str],
    schema: pa.Schema,
    storage_options: dict[str, Any] | None = None,
) -> pl.DataFrame:
    with ThreadPoolExecutor() as executor:
        futures = [
            executor.submit(
                pl.read_parquet,
                file,
                retries=READ_PARQUET_RETRY_COUNT,
                storage_options=storage_options,
                # Hive partitioning was disabled by default in https://github.com/pola-rs/polars/pull/17106
                hive_partitioning=True,
            )
            for file in files
        ]
        concurrent.futures.wait(futures)

    dfs = [future.result() for future in futures]

    return pl.concat([_normalize_df(df, schema=schema) for df in dfs])


def _empty_normalized_df(schema: pa.Schema) -> pl.DataFrame:
    return _normalize_df(pl.DataFrame({col: [] for col in schema.names}), schema=schema)


def _normalize_df(
    df: pl.DataFrame,
    schema: pa.Schema,
    columns: list[str] | None = None,
) -> pl.DataFrame:
    """
    Add missing columns, cast, and reorder dataframe columns to the specified schema's order.

    If columns is provided, only those columns will be added.
    """
    polars_schema = pl.from_arrow(schema.empty_table()).schema
    if columns:
        # Only add back specified columns
        polars_schema = {
            col: dtype for col, dtype in polars_schema.items() if col in columns
        }

    schema_columns = list(polars_schema.keys())
    missing_columns = set(schema_columns) - set(df.columns)
    return (
        df.with_columns(pl.lit(None).alias(col) for col in missing_columns)
        .cast(polars_schema)
        .select(schema_columns)
    )


def datafusion_predicate_from_filters(
    schema: pa.Schema, filters: DeltaInputFilters | None
) -> str | None:
    if not filters:
        return None
    elif isinstance(filters, str):
        return filters

    normalized_filters = normalize_filters(filters)
    return filters_to_sql_predicate(schema, normalized_filters)



from dataclasses import dataclass
from enum import Enum
from typing import Any, NamedTuple, Optional

import boto3
import polars as pl
import pyarrow as pa

from datarepo.core.tables.filters import Filter, NormalizedFilters


@dataclass
class RoapiOptions:
    use_memory_table: bool = False
    disable: bool = False
    override_name: str | None = None
    reload_interval_seconds: int | None = None


@dataclass
class DeltaRoapiOptions(RoapiOptions):
    reload_interval_seconds: int | None = 60


class PartitioningScheme(Enum):
    """
    Defines the partitioning scheme for the table.

    DIRECTORY - e.g. s3://bucket/5956/2024-03-24
    HIVE - e.g. s3://bucket/implant_id=5956/date=2024-03-24
    """

    DIRECTORY = 1
    HIVE = 2


class Partition(NamedTuple):
    column: str
    col_type: pl.DataType


def exactly_one_equality_filter(
    partition: Partition, filters: list[Filter]
) -> Optional[Filter]:
    """Checks whether exactly one equality filter exists for the given partition"""
    match = None

    for f in filters:
        if f.column == partition.column:
            # Multiple matches found, or found comparison that is not equality operator
            if match is not None or f.operator != "=":
                return None
            # First match for full equality operator
            elif f.operator == "=":
                match = f

    return match


def get_storage_options(
    boto3_session: boto3.Session | None = None,
    endpoint_url: str | None = None,
) -> dict[str, str]:
    storage_options = {}

    if endpoint_url is not None:
        storage_options["aws_endpoint_url"] = endpoint_url

    if boto3_session is not None:
        creds = boto3_session.get_credentials()
        storage_options = {
            **storage_options,
            "aws_access_key_id": creds.access_key,
            "aws_secret_access_key": creds.secret_key,
            "aws_session_token": creds.token,
            "aws_region": boto3_session.region_name,
        }

    # Storage options passed to delta-rs need to be not null
    storage_options = {k: v for k, v in storage_options.items() if v}

    return storage_options


def get_pyarrow_filesystem_args(
    boto3_session: boto3.Session | None = None,
    endpoint_url: str | None = None,
) -> dict[str, str]:
    pyarrow_filesystem_args = {}

    if endpoint_url is not None:
        pyarrow_filesystem_args["endpoint_override"] = endpoint_url

    if boto3_session is not None:
        creds = boto3_session.get_credentials()
        pyarrow_filesystem_args = {
            **pyarrow_filesystem_args,
            "access_key": creds.access_key,
            "secret_key": creds.secret_key,
            "session_token": creds.token,
            "region": boto3_session.region_name,
        }

    pyarrow_filesystem_args = {
        k: v for k, v in pyarrow_filesystem_args.items() if v is not None
    }

    return pyarrow_filesystem_args


def filters_to_sql_predicate(schema: pa.Schema, filters: NormalizedFilters) -> str:
    if not filters:
        return "true"

    return " or ".join(
        filters_to_sql_conjunction(schema, filter_set) for filter_set in filters
    )


def filters_to_sql_conjunction(schema: pa.Schema, filters: list[Filter]) -> str:
    if not filters:
        return "true"

    exprs = (filter_to_sql_expr(schema, f) for f in filters)
    conjunction_expr = " and ".join(exprs)
    return f"({conjunction_expr})"


def filter_to_sql_expr(schema: pa.Schema, f: Filter) -> str:
    column = f.column
    if column not in schema.names:
        raise ValueError(f"Invalid column name {column}")

    column_type = schema.field(column).type
    if f.operator in (
        "=",
        "!=",
        "<",
        "<=",
        ">",
        ">=",
        "in",
        "not in",
    ):
        value_str = value_to_sql_expr(f.value, column_type)
        return f"({column} {f.operator} {value_str})"

    elif f.operator == "contains":
        assert isinstance(f.value, str)
        escaped_str = escape_str_for_sql(f.value)
        like_str = f"'%{escaped_str}%'"
        return f"({column} like {like_str})"

    elif f.operator in ("includes", "includes any", "includes all"):
        assert pa.types.is_list(column_type) or pa.types.is_large_list(column_type)

        values: list[Any]
        if f.operator == "includes":
            values = [f.value]
        else:
            assert isinstance(f.value, list | tuple)
            values = list(f.value)

        # NOTE: for includes any/all, we join multiple array_contains with or/and
        value_exprs = (
            value_to_sql_expr(value, column_type.value_type) for value in values
        )
        include_exprs = (
            f"array_contains({column}, {value_expr})" for value_expr in value_exprs
        )
        join_operator = " or " if f.operator == "includes any" else " and "
        conjunction_expr = join_operator.join(include_exprs)

        return f"({conjunction_expr})"

    else:
        raise ValueError(f"Invalid operator {f.operator}")


def value_to_sql_expr(value: Any, value_type: pa.DataType) -> str:
    if isinstance(value, list | tuple):
        elements_str = ", ".join(
            value_to_sql_expr(element, value_type) for element in value
        )
        value_str = f"({elements_str})"
    else:
        value_str = str(value)
        # Escape the string so the user doesn't need to filter like ("col", "=", "'value'")
        if pa.types.is_string(value_type):
            escaped_str = escape_str_for_sql(value_str)
            value_str = f"'{escaped_str}'"
    return value_str


def escape_str_for_sql(value: str) -> str:
    return value.replace("'", "''")



from datarepo.core.catalog import Catalog, Database, ModuleDatabase
from datarepo.core.dataframe import NlkDataFrame
from datarepo.core.tables import (
    DeltaCacheOptions,
    DeltalakeTable,
    Filter,
    ParquetTable,
    Partition,
    PartitioningScheme,
    TableMetadata,
    TableProtocol,
    table,
)

__all__ = [
    "DeltalakeTable",
    "DeltaCacheOptions",
    "ParquetTable",
    "NlkDataFrame",
    "Catalog",
    "table",
    "PartitioningScheme",
    "Partition",
    "Filter",
    "TableMetadata",
    "TableProtocol",
]



from types import ModuleType
from typing import Any, Protocol, cast
import warnings

from datarepo.core.dataframe import NlkDataFrame
from datarepo.core.tables.metadata import TableProtocol


class Database(Protocol):
    def get_tables(self, show_deprecated: bool = False) -> dict[str, TableProtocol]: ...

    def tables(self, show_deprecated: bool = False) -> list[str]:
        return list(self.get_tables(show_deprecated).keys())

    def table(self, name: str, *args: Any, **kwargs: Any) -> NlkDataFrame: ...


class ModuleDatabase(Database):
    def __init__(self, db: ModuleType) -> None:
        self.db = db

    def __getattr__(self, name: str):
        # HACK: to maintain backwards compability when accessing module attributes
        return self._get_table(name)

    def get_tables(self, show_deprecated: bool = False) -> dict[str, TableProtocol]:
        methods = dir(self.db)

        tables = {}
        for name in methods:
            table = self._get_table(name)
            if table is None:
                continue

            if table.table_metadata.is_deprecated and not show_deprecated:
                continue

            tables[name] = table

        return tables

    def table(self, name: str, *args: Any, **kwargs: Any) -> NlkDataFrame:
        tbl = self._get_table(name)
        if tbl is None:
            raise KeyError(f"Table '{name}' not found in database")

        if tbl.table_metadata.is_deprecated:
            warnings.warn(f"The table '{name}' is deprecated", DeprecationWarning)

        return tbl(*args, **kwargs)

    def _get_table(self, name: str) -> TableProtocol | None:
        table = getattr(self.db, name)
        if not hasattr(table, "table_metadata"):
            return None

        return cast(TableProtocol, table)


class DatabaseWithGlobalArgs(Database):
    def __init__(self, db: Database, global_args: dict[str, Any]):
        self.db = db
        self._global_args = global_args

    def get_tables(self, show_deprecated: bool = False) -> dict[str, TableProtocol]:
        return self.db.get_tables(show_deprecated)

    def tables(self, show_deprecated: bool = False) -> list[str]:
        return self.db.tables(show_deprecated)

    def table(self, name: str, *args: Any, **kwargs: Any) -> NlkDataFrame:
        new_kwargs = {**self._global_args, **kwargs}
        return self.db.table(name, *args, **new_kwargs)


class Catalog:
    def __init__(self, dbs: dict[str, Database]):
        self._dbs = dbs
        self._global_args: dict[str, Any] | None = None

    def set_global_args(self, global_args: dict[str, Any]) -> None:
        self._global_args = global_args

    def db(self, db_name: str) -> Database:
        db = self._dbs.get(db_name)

        if db is None:
            raise KeyError(
                f"Database '{db_name}' not found. Available databases: {self.dbs()}"
            )

        if self._global_args is None:
            return db

        return DatabaseWithGlobalArgs(db, self._global_args)

    def dbs(self) -> list[str]:
        return list(self._dbs.keys())



# mypy: disable-error-code=override

from os import path
from typing import Any, Optional, Sequence

import boto3
import polars as pl

from datarepo.core.dataframe import NlkDataFrame
from datarepo.core.tables.filters import (
    InputFilters,
    NormalizedFilters,
    normalize_filters,
)
from datarepo.core.tables.metadata import (
    TableMetadata,
    TableProtocol,
    TableSchema,
)
from datarepo.core.tables.util import (
    Filter,
    Partition,
    PartitioningScheme,
    RoapiOptions,
    exactly_one_equality_filter,
    get_storage_options,
)


def pl_all(exprs: Sequence[pl.Expr]) -> pl.Expr:
    assert len(exprs) > 0

    result = exprs[0]
    for expr in exprs[1:]:
        result = result & expr

    return result


def pl_any(exprs: Sequence[pl.Expr]) -> pl.Expr:
    assert len(exprs) > 0

    result = exprs[0]
    for expr in exprs[1:]:
        result = result | expr

    return result


def _filters_to_conjunction_expr(filters: list[Filter]) -> pl.Expr | None:
    if not filters:
        return None

    result = _filter_to_expr(filters[0])
    for filter in filters[1:]:
        result = result & _filter_to_expr(filter)

    return pl_all([_filter_to_expr(filter) for filter in filters])


def _filters_to_expr(filters: NormalizedFilters) -> pl.Expr | None:
    conjunctions = [_filters_to_conjunction_expr(filter_set) for filter_set in filters]
    not_none_conjunctions = [c for c in conjunctions if c is not None]
    if not not_none_conjunctions:
        return None

    return pl_any(not_none_conjunctions)


def _filter_to_expr(filter: Filter) -> pl.Expr:
    if filter.operator == "=":
        return pl.col(filter.column) == filter.value
    elif filter.operator == "!=":
        return pl.col(filter.column) != filter.value
    elif filter.operator == "<":
        return pl.col(filter.column) < filter.value
    elif filter.operator == ">":
        return pl.col(filter.column) > filter.value
    elif filter.operator == "<=":
        return pl.col(filter.column) <= filter.value
    elif filter.operator == ">=":
        return pl.col(filter.column) >= filter.value
    elif filter.operator == "in":
        return pl.col(filter.column).is_in(filter.value)
    elif filter.operator == "not in":
        return ~pl.col(filter.column).is_in(filter.value)
    elif filter.operator == "contains":
        return pl.col(filter.column).str.contains(filter.value)
    elif filter.operator == "includes":
        return pl.col(filter.column).list.contains(filter.value)
    elif filter.operator == "includes any":
        return pl_any(
            [pl.col(filter.column).list.contains(value) for value in filter.value]
        )
    elif filter.operator == "includes all":
        return pl_all(
            [pl.col(filter.column).list.contains(value) for value in filter.value]
        )
    else:
        raise ValueError(f"Unsupported operator {filter.operator}")


class ParquetTable(TableProtocol):
    def __init__(
        self,
        name: str,
        uri: str,
        partitioning: list[Partition],
        partitioning_scheme: PartitioningScheme = PartitioningScheme.DIRECTORY,
        description: str = "",
        docs_filters: list[Filter] = [],
        docs_columns: list[str] | None = None,
        roapi_opts: RoapiOptions | None = None,
        parquet_file_name: str = "df.parquet",
        table_metadata_args: dict[str, Any] | None = None,
    ):
        if not isinstance(partitioning_scheme, PartitioningScheme):
            raise ValueError(f"Invalid parititioning scheme, got {partitioning_scheme}")

        self.name = name
        self.uri = uri
        self.partitioning = partitioning
        self.partitioning_scheme = partitioning_scheme

        self.table_metadata = TableMetadata(
            table_type="PARQUET",
            description=description,
            docs_args={"filters": docs_filters, "columns": docs_columns},
            roapi_opts=roapi_opts,
            **(table_metadata_args or {}),
        )

        self.parquet_file_name = parquet_file_name

    def get_schema(self) -> TableSchema:
        partitions = [
            {
                "column_name": filter.column,
                "type_annotation": type(filter.value).__name__,
                "value": filter.value,
            }
            for filter in self.table_metadata.docs_args.get("filters", [])
        ]

        # Pop the selected columns so that we still pull the full schema below
        docs_args = {**self.table_metadata.docs_args}
        docs_args.pop("columns")

        columns = None
        if docs_args or not partitions:
            table: NlkDataFrame = self(**docs_args)
            columns = [
                {
                    "name": key,
                    "type": type.__str__(),
                }
                for key, type in table.schema.items()
            ]

        return TableSchema(partitions=partitions, columns=columns)

    def __call__(
        self,
        filters: InputFilters | None = None,
        columns: Optional[list[str]] = None,
        boto3_session: boto3.Session | None = None,
        endpoint_url: str | None = None,
        **kwargs: Any,
    ) -> NlkDataFrame:
        normalized_filters = normalize_filters(filters)
        uri, remaining_partitions, remaining_filters, applied_filters = (
            self._build_uri_from_filters(normalized_filters)
        )

        storage_options = get_storage_options(
            boto3_session=boto3_session,
            endpoint_url=endpoint_url,
        )

        df = pl.scan_parquet(
            uri,
            hive_partitioning=len(remaining_partitions) > 0,
            hive_schema={
                partition.column: partition.col_type
                for partition in remaining_partitions
            },
            allow_missing_columns=True,
            storage_options=storage_options,
        )

        if applied_filters:
            # Add columns removed from partitions and added to uri
            df = df.with_columns(
                pl.lit(f.value)
                .cast(
                    next(
                        partition.col_type
                        for partition in self.partitioning
                        if partition.column == f.column
                    )
                )
                .alias(f.column)
                for f in applied_filters
            )

        if remaining_filters:
            filter_expr = _filters_to_expr(remaining_filters)
            if filter_expr is not None:
                df = df.filter(filter_expr)

        if columns:
            df = df.select(columns)

        return df

    def build_file_fragment(self, filters: list[Filter]) -> str:
        """
        Returns a file path from the base table URI with the given filters.
        This will raise an error if the filter does not specify all partitions.

        This is currently used to generate the file path used by ROAPI to infer schemas.
        """
        uri_with_prefix, partitions, _, _ = self._build_uri_from_filters(
            normalize_filters(filters), include_base_uri=False
        )
        if len(partitions) > 0:
            partition_names = [partition.column for partition in partitions]
            raise ValueError(
                f"Not enough partitions specified, missing: {partition_names}"
            )

        return path.join(uri_with_prefix, self.parquet_file_name)

    def _build_uri_from_filters(
        self,
        filters: NormalizedFilters,
        include_base_uri: bool = True,
    ) -> tuple[str, list[Partition], NormalizedFilters, list[Filter]]:
        """Attempts to build an S3 list prefix from the given filters.
        We do this because pyarrow runs an S3 List() query on the base URI
        before applying filters to possible files. This can be slow.

        We can force pyarrow to do more sophisticated prefix filtering
        if we pre-construct the URI before making the call. If a partition
        has exactly one filter that uses strict equality, we know that
        it will contain that filter string in the URI. We attempt to do this
        for each partition filter, in order, until we encounter a
        partition that:
            1. does not have a filter, or
            2. has a more than one filters, or
            3. has filters that are not strict equality checks.

        This gives us the longest URL that can be used as a prefix filter for
        all files returned by the S3 List() call.

        In a benchmark, this brought reading ~1 million rows of binned spike files
        from 12s to 1.5s.

        In the future, we can further optimize this by building a list of
        candidate URIs and doing separate prefix filtering with each of those, in parallel.

        Long-term we should push this logic into pyarrow and make an upstream commit.

        NOTE: Add trailing slash - this is important to ensure that,
        when partitioning only by implant ID, a future 5-digit implant beginning with the same 4 digits
        as a 4-digit implant is not included in a 4-digit implant's query.
        """
        uri = self.uri if include_base_uri else ""

        if not filters or not self.partitioning:
            return uri, self.partitioning, filters, []

        partitions = list(self.partitioning)
        filters = [list(filter_set) for filter_set in filters]
        applied_filters = []

        for i, partition in enumerate(self.partitioning):
            partition_filters = [
                exactly_one_equality_filter(partition, f) for f in filters
            ]

            # either 0 or multiple filters for this partition,
            # break and deal with the s3 list() query
            if any(partition_filter is None for partition_filter in partition_filters):
                break

            # Only move forward if all remaining filter sets have the same partition filter
            if not all(
                partition_filter == partition_filters[0]
                for partition_filter in partition_filters
            ):
                break

            partition_filter = partition_filters[0]

            if self.partitioning_scheme == PartitioningScheme.DIRECTORY:
                partition_component = str(partition_filter.value)
            elif self.partitioning_scheme == PartitioningScheme.HIVE:
                partition_component = (
                    partition.column + "=" + str(partition_filter.value)
                )

            uri = path.join(uri, partition_component)

            # remove the partition and filter since it's been applied
            # technically might not need to remove the partitions,
            # but they are semantically meaningless as we already have
            # constructed the URI, so we can pop them
            partitions.remove(partition)
            for filter_set in filters:
                filter_set.remove(partition_filter)

            applied_filters.append(partition_filter)

        uri = path.join(
            uri, ""
        )  # trailing slash prevents inclusion of partitions that are subsets of other partitions

        return (uri, partitions, filters, applied_filters)



import inspect
from typing import Any, Callable, TypeVar

from datarepo.core.dataframe.frame import NlkDataFrame
from datarepo.core.tables.metadata import (
    TableMetadata,
    TableProtocol,
    TableSchema,
)

U = TypeVar("U")


class FunctionTable(TableProtocol):
    def __init__(self, table_metadata: TableMetadata, func: Callable) -> None:
        self.table_metadata = table_metadata
        self.func = func

    def __call__(self, *args: Any, **kwargs: dict[str, Any]) -> NlkDataFrame:
        # Filter to only include kwargs that are in the function signature
        parameters = inspect.signature(self.func).parameters
        accepts_var_kwargs = any(
            param.kind == inspect.Parameter.VAR_KEYWORD for param in parameters.values()
        )
        if not accepts_var_kwargs:
            kwargs = {
                key: value
                for key, value in kwargs.items()
                if key in parameters and key not in args
            }

        return self.func(*args, **kwargs)

    def get_schema(self) -> TableSchema:
        filters = self.table_metadata.docs_args.get("filters", [])

        # Infer partitions from filters
        partitions = [
            {
                "column_name": filter.column,
                "type_annotation": type(filter.value).__name__,
                "value": filter.value,
            }
            for filter in filters
        ]

        columns = None
        if self.table_metadata.docs_args or not partitions:
            fallback_table = self(**self.table_metadata.docs_args)
            columns = [
                {
                    "name": key,
                    "type": type.__str__(),
                }
                for key, type in fallback_table.collect_schema().items()
            ]

        return TableSchema(partitions=partitions, columns=columns)


def table(*args, **kwargs) -> Callable[[U], U] | Callable[[Any], Callable[[U], U]]:
    def wrapper(func):
        return FunctionTable(
            table_metadata=TableMetadata(
                table_type="FUNCTION",
                description=func.__doc__.strip() if func.__doc__ else "",
                docs_args=kwargs.get("docs_args", {}),
                latency_info=kwargs.get("latency_info"),
                example_notebook=kwargs.get("example_notebook"),
                data_input=kwargs.get("data_input"),
                is_deprecated=kwargs.get("is_deprecated", False),
            ),
            func=func,
        )

    if len(args) == 0:
        return wrapper
    else:
        return wrapper(args[0])
